from contextlib import nullcontext

import pytest
import torch
from torch import nn

from nvbenjo.cfg import TorchRuntimeConfig
from nvbenjo.torch_utils import (
    apply_batch_precision,
    apply_non_amp_model_precision,
    get_amp_ctxt_for_precision,
    get_model_parameters,
)
from nvbenjo.utils import CompileMode, PrecisionType


def test_get_model_parameters():
    class SimpleModel(nn.Module):
        def __init__(self):
            super(SimpleModel, self).__init__()
            self.fc = nn.Linear(10, 10, bias=False)

        def forward(self, x):
            return self.fc(x)

    model = SimpleModel()
    num_params = get_model_parameters(model)
    assert num_params == 100


def test_apply_non_amp_model_precision():
    class SimpleModel(nn.Module):
        def __init__(self):
            super(SimpleModel, self).__init__()
            self.fc = nn.Linear(10, 10, bias=False)

        def forward(self, x):
            return self.fc(x)

    model = SimpleModel()
    batch = torch.randn(10, 10)
    model = apply_non_amp_model_precision(model, PrecisionType.FP16)
    batch = apply_batch_precision(batch, PrecisionType.FP16)
    assert model.fc.weight.dtype == torch.float16
    assert batch.dtype == torch.float16

    model = SimpleModel()
    batch = torch.randn(10, 10)
    model = apply_non_amp_model_precision(model, PrecisionType.FP32)
    batch = apply_batch_precision(batch, PrecisionType.FP32)
    assert model.fc.weight.dtype == torch.float32
    assert batch.dtype == torch.float32

    model = SimpleModel()
    batch = torch.randn(10, 10)
    model = apply_non_amp_model_precision(model, PrecisionType.BFLOAT16)
    batch = apply_batch_precision(batch, PrecisionType.BFLOAT16)
    assert model.fc.weight.dtype == torch.bfloat16
    assert batch.dtype == torch.bfloat16

    model = SimpleModel()
    batch = torch.randn(10, 10)
    model = apply_non_amp_model_precision(model, PrecisionType.AMP_FP16)
    batch = apply_batch_precision(batch, PrecisionType.AMP_FP16)
    # only shall apply non-amp precisions
    assert model.fc.weight.dtype == torch.float32
    assert batch.dtype == torch.float32


def test_get_amp_ctxt_for_precision():
    ctxt = get_amp_ctxt_for_precision(PrecisionType.AMP, torch.device("cpu"))
    assert isinstance(ctxt, torch.autocast)

    ctxt = get_amp_ctxt_for_precision(PrecisionType.FP32, torch.device("cpu"))
    assert isinstance(ctxt, nullcontext)


@pytest.mark.parametrize(
    "compile_input,expected_mode",
    [
        (False, CompileMode.NONE),
        (True, CompileMode.TORCH_COMPILE),
        ("torch_compile", CompileMode.TORCH_COMPILE),
        ("aot_compile", CompileMode.AOT_COMPILE),
        ("none", CompileMode.NONE),
        ("AOT_COMPILE", CompileMode.AOT_COMPILE),
    ],
)
def test_runtime_config_compile_mode(compile_input, expected_mode):
    cfg = TorchRuntimeConfig(compile=compile_input)
    assert cfg._compile_mode == expected_mode


def test_runtime_config_compile_invalid():
    with pytest.raises(ValueError):
        TorchRuntimeConfig(compile="invalid_mode")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

# Add the src directory to the path so Sphinx can find nvbenjo
sys.path.insert(0, os.path.abspath("../src"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Nvbenjo"
copyright = "2026, lukas-jkl"
author = "lukas-jkl"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",  # Auto-generate docs from docstrings
    "sphinx.ext.napoleon",  # Support for NumPy/Google style docstrings
    "sphinx.ext.viewcode",  # Add links to source code
    "sphinx.ext.intersphinx",  # Link to other projects' documentation
    "myst_parser",
    "sphinx_copybutton",
]
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
# Napoleon settings for better NumPy-style formatting
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = True
napoleon_type_aliases = {
    "TensorLike": "nvbenjo.utils.TensorLike",
    "array_like": ":term:`array_like`",
}
napoleon_attr_annotations = True

# Add NumPy intersphinx for better links
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "torch": ("https://pytorch.org/docs/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
}
# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "shibuya"
html_static_path = ["_static"]

# Shibuya theme options
html_theme_options = {
    "github_url": "https://github.com/lukas-jkl/nvbenjo",
    "accent_color": "indigo",
}
html_show_sourcelink = False
html_copy_source = False

# -- Options for autodoc extension -------------------------------------------
autodoc_default_options = {
    "member-order": "bysource",
    "members": False,
    "undoc-members": False,
    "exclude-members": "__weakref__, __init__",
    "inherited-members": False,
}

# Show type hints in signature only
autodoc_typehints = "description"
autodoc_typehints_format = "short"
autodoc_class_signature = "mixed"



import logging
import os
import time
import typing as ty
from collections.abc import Sequence
from contextlib import AbstractContextManager, nullcontext
from typing import Callable, Optional

import pandas as pd
import torch
import torch.nn as nn
import torchvision

from nvbenjo import console
from nvbenjo.cfg import TorchRuntimeConfig
from nvbenjo.utils import AMP_PREFIX, TRANSFER_WARNING, PrecisionType, TensorLike

logger = logging.getLogger(__name__)


def get_model(
    type_or_path: str, device: torch.device, runtime_config: TorchRuntimeConfig, verbose=False, **kwargs
) -> ty.Any:
    """Load PyTorch model.

    Parameters
    ----------
    type_or_path : str
        Model type or path. Supports prefixes to specify the model source:

        - ``torchvision:<name>`` -- Load a torchvision model (e.g. ``torchvision:resnet50``), see `torchvision.models.list_models()`
        - ``huggingface:<name>`` -- Load a HuggingFace AutoModel (e.g. ``huggingface:bert-base-uncased``), see https://huggingface.co/docs/transformers/model_doc/auto
        - ``jit:<path>`` -- Load a TorchScript/JIT model
        - ``torchexport:<path>`` -- Load a ``torch.export`` saved model
        - ``aot:<path>`` -- Load a pre-compiled AOT model
        - *(no prefix)* -- Path to a model saved with ``torch.save`` or ``torch.jit.save``

    device : torch.device
        Device to load the model onto.
    runtime_config : TorchRuntimeConfig
        Runtime configuration for the model.
    verbose : bool, optional
        Whether to print verbose output, by default False

    Returns
    -------
    ty.Any
        Loaded model.

    Examples
    --------
    >>> model = get_model("torchvision:resnet18", device=torch.device("cpu"), runtime_config=TorchRuntimeConfig())
    >>> model = get_model("/path/to/model.pth", device=torch.device("cuda"), runtime_config=TorchRuntimeConfig())
    >>> model = get_model("jit:/path/to/model.pt", device=torch.device("cuda"), runtime_config=TorchRuntimeConfig())
    >>> model = get_model("torchexport:/path/to/model.pt2", device=torch.device("cuda"), runtime_config=TorchRuntimeConfig())
    >>> model = get_model("aot:/path/to/model.pt2", device=torch.device("cuda"), runtime_config=TorchRuntimeConfig())
    >>> model = get_model("huggingface:bert-base-uncased", device=torch.device("cpu"), runtime_config=TorchRuntimeConfig())
    """
    type_or_path = os.path.expanduser(type_or_path)
    if type_or_path.startswith("jit:"):
        if verbose and console is not None:
            console.print(f"Loading jit model {type_or_path}")
        type_or_path = type_or_path[len("jit:") :]
        return torch.jit.load(type_or_path, map_location=device)
    elif type_or_path.startswith("torchexport:"):
        if verbose and console is not None:
            console.print(f"Loading torchexport model {type_or_path}")
        type_or_path = type_or_path[len("torchexport:") :]
        program = torch.export.load(type_or_path)
        module = program.module()
        module = module.to(device)
        return module
    elif type_or_path.startswith("aot:"):
        if verbose and console is not None:
            console.print(f"Loading AOT model {type_or_path}")
        type_or_path = type_or_path[len("aot:") :]
        return torch._inductor.aoti_load_package(type_or_path)
    elif os.path.isfile(type_or_path):
        # Path and no prefix -> try different methods
        if verbose and console is not None:
            console.print(f"Loading torch model {type_or_path}")
        try:
            model = torch.load(type_or_path, map_location=device, weights_only=False)
            model.eval()
            return model
        except Exception:
            try:
                return torch.jit.load(type_or_path, map_location=device)
            except Exception:
                if torch.__version__ > "2.1":
                    try:
                        program = torch.export.load(type_or_path)
                        module = program.module()
                        module = module.to(device)
                        return module
                    except Exception:
                        return torch._inductor.aoti_load_package(type_or_path)
                else:
                    raise

    if type_or_path.startswith("huggingface:"):
        type_or_path = type_or_path[len("huggingface:") :]
        if verbose and console is not None:
            console.print(f"Loading huggingface model {type_or_path}")
        from transformers import AutoModel  # type: ignore

        return AutoModel.from_pretrained(type_or_path).to(device)
    elif type_or_path.startswith("torchvision:"):
        type_or_path = type_or_path[len("torchvision:") :]
        available_torchvision_models = torchvision.models.list_models()
        if type_or_path in available_torchvision_models:
            if verbose and console is not None:
                console.print(f"Loading torchvision model {type_or_path}")
            model = torchvision.models.get_model(type_or_path, **kwargs).to(device)
            model.eval()
            return model
    else:
        available_torchvision_models = torchvision.models.list_models()
        raise ValueError(
            f"Invalid model {type_or_path}. Must be: \n"
            "- a valid path to a saved torch model\n"
            "- 'jit:<path>' for a TorchScript/JIT model\n"
            "- 'torchexport:<path>' for a torch.export model\n"
            "- 'aot:<path>' for a pre-compiled AOT model\n"
            "- 'huggingface:<model-name>' for a HuggingFace AutoModel\n"
            f"- 'torchvision:<model-name>' from {available_torchvision_models}\n"
        )


def run_model_with_input(model: nn.Module | Callable, input: TensorLike) -> TensorLike:
    if isinstance(input, (list, tuple)):
        return model(*input)
    elif isinstance(input, dict):
        return model(**{str(k): v for k, v in input.items()})
    else:
        return model(input)


def transfer_to_device(result: ty.Any, to_device: torch.device) -> ty.Any:
    if hasattr(result, "to"):
        return result.to(to_device)
    if isinstance(result, Sequence):
        return [transfer_to_device(ri, to_device=to_device) for ri in result]
    elif hasattr(result, "items"):
        return {k: transfer_to_device(v, to_device=to_device) for k, v in result.items()}
    else:
        raise ValueError(f"Unsupported result type: {type(result)} could not transfer to {to_device}")


def apply_batch_precision(batch: TensorLike, precision: PrecisionType) -> TensorLike:
    def _apply_batch_precision(batch_tensor: torch.Tensor):
        if AMP_PREFIX not in precision.value:
            if precision == PrecisionType.FP16:
                batch_tensor = batch_tensor.half()
            elif precision == PrecisionType.BFLOAT16:
                batch_tensor = batch_tensor.bfloat16()
            else:
                if precision != PrecisionType.FP32:
                    raise ValueError(f"Invalid precision type {precision}.")
        return batch_tensor

    if isinstance(batch, torch.Tensor):
        batch = _apply_batch_precision(batch)
    elif isinstance(batch, (list, tuple)):
        batch = tuple(_apply_batch_precision(b) for b in batch)
    elif isinstance(batch, dict):
        batch = {k: _apply_batch_precision(v) for k, v in batch.items()}
    else:
        raise ValueError(f"Unsupported batch type: {type(batch)}. Must be a Tensor, Tuple, or Dict.")

    return batch


def apply_non_amp_model_precision(
    model: nn.Module,
    precision: PrecisionType,
) -> nn.Module:
    if AMP_PREFIX not in precision.value:
        if precision == PrecisionType.FP16:
            model = model.half()
        elif precision == PrecisionType.BFLOAT16:
            model = model.bfloat16()
        else:
            if precision != PrecisionType.FP32:
                raise ValueError(f"Invalid precision type {precision}.")

    return model


def get_amp_ctxt_for_precision(precision: PrecisionType, device: torch.device) -> AbstractContextManager:
    if AMP_PREFIX in precision.value:
        valid_values = [PrecisionType.AMP, PrecisionType.AMP_FP16, PrecisionType.AMP_BFLOAT16]
        if precision not in valid_values:
            raise ValueError(f"Invalid AMP precision type {precision} must be one of {valid_values}")

        if precision in [PrecisionType.AMP]:
            ctxt = torch.autocast(device_type=device.type, enabled=True)
        elif precision in [PrecisionType.AMP_FP16]:
            ctxt = torch.autocast(device_type=device.type, dtype=torch.float16, enabled=True)
        elif precision in [PrecisionType.AMP_BFLOAT16]:
            ctxt = torch.autocast(device_type=device.type, dtype=torch.bfloat16, enabled=True)
        else:
            raise ValueError(f"Invalid precision type {precision}.")
    else:
        ctxt = nullcontext()
    return ctxt


def get_model_parameters(model: nn.Module) -> int:
    return sum(p.numel() for p in model.parameters())


def measure_memory_allocation(
    model: nn.Module | Callable, batch: TensorLike, device: torch.device, iterations: int = 3
) -> int:
    """Measure the peak memory usage during inference

    Parameters
    ----------
    model : nn.Module | Callable
        The model to benchmark.
    batch : TensorLike
        Sample input to the model.
    device : torch.device
        The device where the model is located and shall be used for benchmarking.
    iterations : int, optional
        Number of iterations to run for measuring memory allocation, by default 3

    Returns
    -------
    int
        Maximum memory allocated during inference in bytes.
    """
    if device.type == "cuda":
        torch.cuda.reset_peak_memory_stats(device=device)
    # before_run_allocation = torch.cuda.memory_allocated(device=device)

    batch = transfer_to_device(batch, to_device=device)
    if isinstance(model, nn.Module):
        model = model.to(device)
    for _ in range(iterations):
        r = run_model_with_input(model, batch)
    try:
        _ = transfer_to_device(r, to_device=torch.device("cpu"))
    except Exception:
        console.print(TRANSFER_WARNING)

    if device.type == "cuda":
        logger.debug(torch.cuda.memory_summary(device=device, abbreviated=True))
        # after_batch_allocation = torch.cuda.memory_allocated(device=device)
        max_batch_allocation = torch.cuda.max_memory_allocated(device=device)
    else:
        max_batch_allocation = -1

    return max_batch_allocation


def measure_repeated_inference_timing(
    model: nn.Module,
    sample: TensorLike,
    batch_size: int,
    model_device: torch.device,
    transfer_to_device_fn: Callable = transfer_to_device,
    num_runs: int = 100,
    progress_callback: Optional[Callable] = None,
) -> pd.DataFrame:
    """Measure inference times.

    Parameters
    ----------
    model : nn.Module
        The model to benchmark.
    sample : TensorLike
        Sample input to the model.
    batch_size : int
        The batch size of the sample.
    model_device : torch.device
        The device where the model is located and shall be used for benchmarking.
    transfer_to_device_fn : Callable, optional
        Function to transfer data to the specified device, by default transfer_to_device
    num_runs : int, optional
        Number of inference runs to perform, by default 100
    progress_callback : Optional[Callable], optional
        Callback function to report progress, by default None

    Returns
    -------
    pd.DataFrame
        DataFrame containing timing results.

    Examples
    --------
    Measure Inference::

        import torch
        from nvbenjo.torch_utils import measure_repeated_inference_timing
        from nvbenjo.torch_utils import get_model
        from nvbenjo.cfg import TorchRuntimeConfig

        model = get_model("torchvision:resnet18", device=torch.device("cpu"), runtime_config=TorchRuntimeConfig())
        sample = torch.randn(2, 3, 224, 224)  # batch size 2
        results = measure_repeated_inference_timing(
            model=model,
            sample=sample,
            batch_size=2,
            model_device=torch.device("cpu"),
            num_runs=2
        )

    """
    results_raw = []

    for _ in range(num_runs):
        start_on_cpu = time.perf_counter()
        device_sample = transfer_to_device_fn(sample, model_device)

        if model_device.type == "cuda":
            start_event = torch.cuda.Event(enable_timing=True)
            stop_event = torch.cuda.Event(enable_timing=True)
            start_event.record()  # For GPU timing
        start_on_device = time.perf_counter()  # For CPU timing

        device_result = run_model_with_input(model, device_sample)

        if model_device.type == "cuda":
            stop_event.record()
            torch.cuda.synchronize()
            # elapsed_on_device = stop_event.elapsed_time(start_event)
            elapsed_on_device = start_event.elapsed_time(stop_event) / 1000.0
            stop_on_device = time.perf_counter()
        else:
            stop_on_device = time.perf_counter()
            elapsed_on_device = stop_on_device - start_on_device

        try:
            transfer_to_device_fn(device_result, torch.device("cpu"))
        except Exception:
            console.print(TRANSFER_WARNING)
        stop_on_cpu = time.perf_counter()

        assert elapsed_on_device > 0

        results_raw.append(
            {
                "time_cpu_to_device": start_on_device - start_on_cpu,
                "time_inference": elapsed_on_device,
                "time_device_to_cpu": stop_on_cpu - stop_on_device,
                "time_total": stop_on_cpu - start_on_cpu,
                "time_total_batch_normalized": (stop_on_cpu - start_on_cpu) / batch_size,
            }
        )
        if progress_callback is not None:
            progress_callback()

    results_raw = pd.DataFrame(results_raw)

    return results_raw



from __future__ import annotations

from platform import uname
from typing import Any
import logging

import pynvml
from cpuinfo import get_cpu_info
import psutil

from .utils import format_num

logger = logging.getLogger(__name__)


def _get_architecture_name_from_version(version: int) -> str:
    version_names = {
        pynvml.NVML_DEVICE_ARCH_KEPLER: "Kepler",
        pynvml.NVML_DEVICE_ARCH_MAXWELL: "Maxwell",
        pynvml.NVML_DEVICE_ARCH_PASCAL: "Pascal",
        pynvml.NVML_DEVICE_ARCH_VOLTA: "Volta",
        pynvml.NVML_DEVICE_ARCH_TURING: "Truing",
        pynvml.NVML_DEVICE_ARCH_AMPERE: "Ampere",
        pynvml.NVML_DEVICE_ARCH_ADA: "Ada",
        pynvml.NVML_DEVICE_ARCH_HOPPER: "Hopper",
        pynvml.NVML_DEVICE_ARCH_BLACKWELL: "Blackwell",
    }
    return f"Version {version} ({version_names.get(version, 'Unknown')})"


def get_gpu_info() -> list[dict[str, Any]]:
    """Retrieve information about GPUs in the system.

    Includes information such as name, architecture, memory, clock speeds, CUDA capability, and driver version.

    Returns
    -------
    list[dict[str, Any]]
        A list of dictionaries containing GPU information.
    """
    pynvml.nvmlInit()
    device_count = pynvml.nvmlDeviceGetCount()
    infos = []
    for i in range(device_count):
        handle = pynvml.nvmlDeviceGetHandleByIndex(i)
        compute_capability = pynvml.nvmlDeviceGetCudaComputeCapability(handle)
        clock_gpu = pynvml.nvmlDeviceGetClockInfo(handle, pynvml.NVML_CLOCK_GRAPHICS)
        clock_mem = pynvml.nvmlDeviceGetClockInfo(handle, pynvml.NVML_CLOCK_MEM)
        # clock_sm = pynvml.nvmlDeviceGetClockInfo(handle, pynvml.NVML_CLOCK_SM)
        # clock_video = pynvml.nvmlDeviceGetClockInfo(handle, pynvml.NVML_CLOCK_VIDEO)

        infos.append(
            {
                "idx": i,
                "name": pynvml.nvmlDeviceGetName(handle),
                "architecture": _get_architecture_name_from_version(pynvml.nvmlDeviceGetArchitecture(handle)),
                "memory": format_num(pynvml.nvmlDeviceGetMemoryInfo(handle).total, bytes=True),
                "clock_gpu": f"{clock_gpu} Mhz",
                "clock_mem": f"{clock_mem} Mhz",
                "cuda_capability": f"{compute_capability[0]}.{compute_capability[1]}",
                "driver": str(pynvml.nvmlSystemGetDriverVersion()),
            }
        )
    pynvml.nvmlShutdown()
    return infos


def get_gpu_power_usage(device_index: int) -> float:
    pynvml.nvmlInit()
    usage = pynvml.nvmlDeviceGetPowerUsage(pynvml.nvmlDeviceGetHandleByIndex(device_index))
    pynvml.nvmlShutdown()
    return usage


def get_system_info() -> dict[str, Any]:
    """Retrieve system information.

    Collects information about the operating system, CPU, memory, and GPU.

    Returns
    -------
    dict[str, Any]
        A dictionary containing system information.
    """
    sys = uname()
    cpu = get_cpu_info()
    svmem = psutil.virtual_memory()
    try:
        gpus = get_gpu_info()
    except pynvml.NVMLError_LibraryNotFound:  # type: ignore
        logger.warning("NVIDIA driver not found")
        gpus = {}
    if hasattr(psutil, "cpu_freq") and psutil.cpu_freq() is not None:
        cpufreq = psutil.cpu_freq().max
    else:
        cpufreq = 0.0
    return {
        "os": {"system": sys.system, "node": sys.node, "release": sys.release, "version": sys.version},
        "cpu": {
            "model": cpu["brand_raw"],
            "architecture": cpu["arch_string_raw"],
            "cores": {
                "physical": psutil.cpu_count(logical=False),
                "total": psutil.cpu_count(logical=True),
            },
            "frequency": f"{(cpufreq / 1000):.2f} GHz",
        },
        "memory": format_num(svmem.total, bytes=True),
        "gpus": gpus,
    }



import logging
import os
import typing as ty
from abc import ABC
from contextlib import nullcontext
from dataclasses import dataclass, field

from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf, open_dict

from .utils import PrecisionType, ProviderType, CompileMode

logger = logging.getLogger(__name__)


@dataclass
class BaseModelConfig(ABC):
    """Base model configuration

    Parameters
    ----------
    name : str
        Name of the model.
    type_or_path : str
        Model type or path. Can be a local file path or a model identifier.
    kwargs : dict
        Additional keyword arguments to pass when instantiating the model.
    shape : tuple
        Input shape of the model. Use "B" or "batch_size" to denote the batch size dimension.
    num_warmup_batches : int
        Number of warm-up batches to run before measuring performance.
    num_batches : int
        Number of batches to run for performance measurement.
    batch_sizes : tuple
        Tuple of batch sizes to benchmark.
    devices : tuple of str
        Tuple of device names to benchmark on (e.g., 'cpu', 'cuda:0').
    runtime_options : dict[str, ty.Any]
        Dictionary mapping runtime names to their specific runtime configurations.
    custom_batchmetrics : dict[str, float]
        Dictionary of custom batch metrics to include in the benchmark results.
        These are calculated as `value / time_total_batch_normalized`.
        For example if each batch is a frame you can calculate FPS as:
            custom_batchmetrics:
                fps: 1.0
                batch-per-second: 1.0
        or if each batch is 3s of audio you can calculate the real time factor as:
            custom_batchmetrics:
                real-time-factor: 1/3
    """

    name: str = "resnet"
    type_or_path: str = "torchvision:wide_resnet101_2"
    kwargs: dict = field(default_factory=dict)
    shape: tuple = ("B", 3, 224, 224)
    num_warmup_batches: int = 5
    num_batches: int = 50
    batch_sizes: tuple = (16, 32)
    devices: tuple[str] = ("cpu",)
    runtime_options: dict[str, ty.Any] = field(default_factory=dict)
    custom_batchmetrics: dict[str, float] = field(default_factory=dict)


@dataclass
class NvbenjoConfig:
    """
    Root configuration for nvbenjo benchmarking.

    Parameters
    ----------
    measure_memory : bool
        Whether to measure GPU memory allocation during benchmarking.
    models: dict[str, TorchModelConfig | OnnxModelConfig]
        Dictionary mapping model names to their configurations.
        See :class:`TorchModelConfig` and :class:`OnnxModelConfig` for details.
    """

    measure_memory: bool = True
    models: dict[str, ty.Any] = field(default_factory=lambda: dict())


@dataclass
class BenchConfig:
    """
    Main benchmark configuration container.

    Parameters
    ----------
    nvbenjo : NvbenjoConfig
        Nvbenjo-specific configuration settings.
    output_dir : str or None
        Directory path where benchmark results will be saved.
        If None, uses Hydra's default output directory.
    """

    nvbenjo: NvbenjoConfig = field(default_factory=NvbenjoConfig)
    output_dir: ty.Optional[str] = None


@dataclass
class TorchRuntimeConfig:
    """PyTorch Runtime configuration:

    Parameters
    ----------
    compile : str
        Model compilation mode:

        - ``false`` -- No compilation (default)
        - ``torch_compile`` -- Compile with ``torch.compile`` (PyTorch 2.0+)
        - ``aot_compile`` -- Ahead-of-time compilation via ``torch._inductor``
    compile_kwargs : dict
        Additional keyword arguments passed to ``torch.compile`` or ``aoti_compile_and_package``.
    precision : PrecisionType
        Precision type for model inference (e.g., fp32, fp16, amp).
    enable_profiling : bool
        Whether to enable PyTorch profiler during inference.
    profiling_prefix : str or None
        Prefix for profiler output files. If None, a default path will be used.
    profiler_kwargs : dict
        Additional keyword arguments for torch.profiler.profile.
    """

    compile: str = "False"
    compile_kwargs: dict = field(default_factory=dict)
    precision: PrecisionType = PrecisionType.FP32
    enable_profiling: bool = False
    profiling_prefix: ty.Optional[str] = None
    profiler_kwargs: dict = field(default_factory=dict)

    def __post_init__(self):
        # Hydra passes everything as primitives, normalize here
        if isinstance(self.compile, bool) or str(self.compile).lower() in ("true", "false"):
            is_true = self.compile is True or str(self.compile).lower() == "true"
            self.compile = "torch_compile" if is_true else "none"
        self._compile_mode = CompileMode(self.compile.lower())


@dataclass
class OnnxRuntimeConfig:
    """ONNX Runtime configuration:

    Parameters
    ----------
    execution_providers : tuple of str or None
        Tuple of execution providers to use (e.g., ('CPUExecutionProvider',
        'CUDAExecutionProvider')). If None, uses the default provider.
    graph_optimization_level : str
        Graph optimization level for ONNX Runtime. Options are 'ORT_ENABLE_ALL', 'ORT_ENABLE_LAYOUT',
        'ORT_ENABLE_BASIC', 'ORT_DISABLE_ALL'.
    intra_op_num_threads : int
        Number of threads used to parallelize the execution within nodes.
    inter_op_num_threads : int
        Number of threads used to parallelize the execution of the graph (between nodes)
    log_severity_level : int
        Logging severity level (0=VERBOSE, 1=INFO, 2=WARNING, 3=ERROR, 4=FATAL)
    enable_profiling : bool
        Whether to enable profiling in ONNX Runtime.
    profiling_prefix : str or None
        Prefix for profiling output files. If None, a default path will be used.
    provider_options : sequence of dict or None
        Additional options for each execution provider.
    """

    execution_providers: ty.Optional[ty.List[ProviderType]] = None
    graph_optimization_level: str = (
        "ORT_ENABLE_ALL"  # 99 ORT_ENABLE_ALL, 3 ORT_ENABLE_LAYOUT, 1 ORT_ENABLE_BASIC, 0 ORT_DISABLE_ALL
    )
    intra_op_num_threads: int = 1
    inter_op_num_threads: int = 0
    log_severity_level: int = 3  # Error
    enable_profiling: bool = False
    profiling_prefix: ty.Optional[str] = None
    provider_options: ty.Sequence[dict[ty.Any, ty.Any]] | None = None


@dataclass
class TorchModelConfig(BaseModelConfig):
    """PyTorch model configuration

    Parameters
    ----------
    name : str
        Name of the model.
    type_or_path : str
        Model type or path. Supports prefixes to specify the model source:

        - ``torchvision:<name>`` -- Load a torchvision model (e.g. ``torchvision:resnet50``)
        - ``huggingface:<name>`` -- Load a HuggingFace AutoModel (e.g. ``huggingface:bert-base-uncased``)
        - ``jit:<path>`` -- Load a TorchScript/JIT model
        - ``torchexport:<path>`` -- Load a ``torch.export`` saved model
        - ``aot:<path>`` -- Load a pre-compiled AOT model

        .. note::

            For ``torchexport`` and ``aot`` models, precision is baked in at export time
            and cannot be changed at runtime.

        - *(no prefix)* -- Path to a model saved with ``torch.save`` or ``torch.jit.save``
    kwargs : dict
        Additional keyword arguments to pass when instantiating the model.
    shape : tuple
        Input shape of the model. Use "B" to denote the batch size dimension.
        Examples::

            # Single input shape
            ("B", 3, 224, 224)

            # Multiple input shapes
            (("B", 3, 224, 224), ("B", 10))

            # Dictionary with metadata
            ({"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},)

            # Multiple dictionary inputs
            (
                {"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},
                {"name": "input2", "type": "int", "shape": (1, 3)},
                {"name": "input3", "type": "int", "shape": (), "value": 42},
            )
    num_warmup_batches : int
        Number of warm-up batches to run before measuring performance.
    num_batches : int
        Number of batches to run for performance measurement.
    batch_sizes : tuple
        Tuple of batch sizes to benchmark.
    devices : tuple of str
        Tuple of device names to benchmark on (e.g., 'cpu', 'cuda:0').
    runtime_options : dict[str, :class:`~nvbenjo.cfg.TorchRuntimeConfig`]
        Dictionary mapping runtime names to their specific runtime configurations.
    """

    model_kwargs: dict = field(default_factory=dict)
    runtime_options: dict[str, TorchRuntimeConfig] = field(default_factory=lambda: {"default": TorchRuntimeConfig()})

    def __post_init__(self):
        for i, (key, opt) in enumerate(self.runtime_options.items()):
            if isinstance(opt, DictConfig):
                self.runtime_options[key] = OmegaConf.structured(TorchRuntimeConfig(**OmegaConf.to_container(opt)))  # type: ignore

        if self.type_or_path.startswith(("aot:", "torchexport:")):
            precisions = {rt.precision for rt in self.runtime_options.values() if isinstance(rt, TorchRuntimeConfig)}
            if len(precisions) > 1:
                raise ValueError(
                    f"Model '{self.name}' is pre-exported — precision is baked in at export time. "
                    f"Got multiple different precisions in runtime_options: {precisions}"
                )
            compile_modes = {
                rt._compile_mode for rt in self.runtime_options.values() if isinstance(rt, TorchRuntimeConfig)
            }
            if compile_modes - {CompileMode.NONE}:
                logger.warning(f"Model '{self.name}' is pre-exported — setting compile has no effect.")


@dataclass
class OnnxModelConfig(BaseModelConfig):
    """ONNX model configuration

    Parameters
    ----------
    name : str
        Name of the model.
    type_or_path : str
        Model type or path. Can be a local file path or a model identifier.
    kwargs : dict
        Additional keyword arguments to pass when instantiating the model.
    shape : tuple
        Input shape of the model. Use "B" to denote the batch size dimension.

        Examples::

            # Single input shape
            ("B", 3, 224, 224)

            # Multiple input shapes
            (("B", 3, 224, 224), ("B", 10))

            # Dictionary with metadata
            ({"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},)

            # Multiple dictionary inputs
            (
                {"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},
                {"name": "input2", "type": "int", "shape": (1, 3)},
                {"name": "input3", "type": "int", "shape": (), "value": 42},
            )
    num_warmup_batches : int
        Number of warm-up batches to run before measuring performance.
    num_batches : int
        Number of batches to run for performance measurement.
    batch_sizes : tuple
        Tuple of batch sizes to benchmark.
    devices : tuple of str
        Tuple of device names to benchmark on (e.g., 'cpu', 'cuda:0').
    runtime_options : dict[str, :class:`~nvbenjo.cfg.OnnxRuntimeConfig`]
        Dictionary mapping runtime names to their specific runtime configurations.
    """

    runtime_options: dict[str, OnnxRuntimeConfig] = field(default_factory=lambda: {"default": OnnxRuntimeConfig()})

    def __post_init__(self):
        for i, (key, opt) in enumerate(self.runtime_options.items()):
            if isinstance(opt, DictConfig):
                self.runtime_options[key] = OmegaConf.structured(OnnxRuntimeConfig(**OmegaConf.to_container(opt)))  # type: ignore


def instantiate_model_configs(cfg: ty.Union[BenchConfig, DictConfig]) -> dict[str, BaseModelConfig]:
    models = {}
    runtimes = {}
    for model_name, model in cfg.nvbenjo.models.items():
        ctxt = open_dict(model) if isinstance(model, DictConfig) else nullcontext()
        if "_target_" not in model:
            with ctxt:
                if model["type_or_path"].endswith(".onnx"):
                    cfg.nvbenjo.models[model_name]["_target_"] = (
                        f"{OnnxModelConfig.__module__}.{OnnxModelConfig.__qualname__}"
                    )
                    cfg.nvbenjo.models[model_name]["_convert_"] = "all"
                    if "runtime_options" in model:
                        runtimes[model_name] = {}
                        for runtime_name in model["runtime_options"].keys():
                            cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]["_target_"] = (
                                f"{OnnxRuntimeConfig.__module__}.{OnnxRuntimeConfig.__qualname__}"
                            )
                            runtimes[model_name][runtime_name] = instantiate(
                                cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]
                            )
                else:
                    cfg.nvbenjo.models[model_name]["_target_"] = (
                        f"{TorchModelConfig.__module__}.{TorchModelConfig.__qualname__}"
                    )
                    cfg.nvbenjo.models[model_name]["_convert_"] = "all"
                    if "runtime_options" in model:
                        runtimes[model_name] = {}
                        for runtime_name in model["runtime_options"].keys():
                            cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]["_target_"] = (
                                f"{TorchRuntimeConfig.__module__}.{TorchRuntimeConfig.__qualname__}"
                            )
                            runtimes[model_name][runtime_name] = instantiate(
                                cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]
                            )
                            runtimes[model_name][runtime_name].precision = PrecisionType[
                                cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]["precision"].upper()
                            ]

        models[model_name] = instantiate(model) if isinstance(model, DictConfig) else model
        if model_name in runtimes:
            models[model_name].runtime_options = runtimes[model_name]

    # For onnx profiling we add a valid profiling prefix in the output directory if needed
    for model_name, model in models.items():
        if isinstance(model, (OnnxModelConfig, TorchModelConfig)):
            for runtime_name, runtime in model.runtime_options.items():
                if runtime.enable_profiling:
                    if cfg.output_dir is None:
                        raise ValueError("output_dir must be set when profiling is enabled.")

                    if runtime.profiling_prefix is None:
                        runtime.profiling_prefix = os.path.join(
                            cfg.output_dir, model_name, f"{model_name}_{runtime_name}_profile"
                        )
                    else:
                        # make sure the relative path is inside the output dir
                        if not os.path.abspath(runtime.profiling_prefix) == runtime.profiling_prefix:
                            runtime.profiling_prefix = os.path.abspath(
                                os.path.join(cfg.output_dir, runtime.profiling_prefix)
                            )
                    os.makedirs(os.path.dirname(runtime.profiling_prefix), exist_ok=True)

    return models



from unittest.mock import patch, MagicMock

import pynvml

from nvbenjo.system_info import get_gpu_power_usage, get_system_info


def test_get_system_info():
    system_info = get_system_info()
    assert isinstance(system_info, dict)
    assert "cpu" in system_info
    assert "gpus" in system_info
    assert "os" in system_info

    assert isinstance(system_info["gpus"], (dict, list))
    if isinstance(system_info["gpus"], list):
        for gpu in system_info["gpus"]:
            assert isinstance(gpu, dict)
            assert "name" in gpu
            assert "architecture" in gpu
            assert "memory" in gpu
            assert "cuda_capability" in gpu
            assert "driver" in gpu

    # check basic cpu info
    cpu_info = system_info["cpu"]
    assert isinstance(cpu_info, dict)
    assert "cores" in cpu_info
    assert "frequency" in cpu_info
    assert "model" in cpu_info
    assert "architecture" in cpu_info


@patch("nvbenjo.system_info.pynvml")
def test_get_gpu_power_usage(mock_pynvml):
    mock_pynvml.nvmlDeviceGetPowerUsage.return_value = 150000
    assert get_gpu_power_usage(0) == 150000
    mock_pynvml.nvmlInit.assert_called_once()
    mock_pynvml.nvmlShutdown.assert_called_once()


@patch("nvbenjo.system_info.get_gpu_info", side_effect=pynvml.NVMLError_LibraryNotFound())
def test_get_system_info_no_nvidia_driver(mock_gpu):
    info = get_system_info()
    assert info["gpus"] == {}


@patch("nvbenjo.system_info.psutil")
@patch("nvbenjo.system_info.get_gpu_info", return_value=[])
@patch("nvbenjo.system_info.get_cpu_info", return_value={"brand_raw": "test", "arch_string_raw": "x86"})
def test_get_system_info_no_cpu_freq(mock_cpu, mock_gpu, mock_psutil):
    mock_psutil.virtual_memory.return_value = MagicMock(total=8_000_000_000)
    mock_psutil.cpu_count.return_value = 4
    del mock_psutil.cpu_freq  # simulate missing cpu_freq
    info = get_system_info()
    assert info["cpu"]["frequency"] == "0.00 GHz"



import csv
import os
import shutil
import subprocess
import pandas as pd
import tempfile
import warnings
from copy import copy
from os.path import isfile, join

import omegaconf
import pytest
import torch
import yaml
from hydra import compose, initialize

from nvbenjo.cli import run

DATA_FILE = "out.csv"
EXPECTED_OUTPUT_FILES = [
    "config.yaml",
    DATA_FILE,
]


def run_config(cfg):
    if isinstance(cfg, omegaconf.DictConfig):
        run(cfg)
    else:
        raise ValueError("Config is not a DictConfig instance")


def _check_files(directory, files):
    for file in files:
        assert isfile(join(directory, file)), f"File {file} not found in {directory}"
        if file.endswith(".yaml"):
            with open(join(directory, file), "r") as f:
                try:
                    yaml.safe_load(f)
                except yaml.YAMLError as e:
                    raise AssertionError(f"YAML file {file} is not valid: {e}")
        elif file.endswith(".csv"):
            with open(join(directory, file), "r") as f:
                reader = csv.reader(f)
                header = next(reader)
                assert len(header) > 0, f"CSV file {file} is empty or has no header"
                for row in reader:
                    assert len(row) == len(header), f"Row in CSV file {file} does not match header length"


def _check_run_files(cfg: omegaconf.DictConfig):
    expected_files = copy(EXPECTED_OUTPUT_FILES)
    for model_name in cfg.nvbenjo.models.keys():
        expected_files.append(join(model_name, "time_inference.png"))
        expected_files.append(join(model_name, "time_total_batch_normalized.png"))
        expected_files.append(join(model_name, "time_device_to_cpu.png"))
        expected_files.append(join(model_name, "time_cpu_to_device.png"))
        expected_files.append(join(model_name, "memory_bytes.png"))
    if len(cfg.nvbenjo.models) > 1:
        expected_files.append(join("summary", "time_inference.png"))
        expected_files.append(join("summary", "time_total_batch_normalized.png"))
        expected_files.append(join("summary", "time_device_to_cpu.png"))
        expected_files.append(join("summary", "time_cpu_to_device.png"))
        expected_files.append(join("summary", "memory_bytes.png"))
    _check_files(cfg.output_dir, expected_files)
    for model_name in cfg.nvbenjo.models.keys():
        for runtime_name in cfg.nvbenjo.models[model_name].get("runtime_options", {}).keys():
            if cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name].get("enable_profiling", False):
                profile_prefix = cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name].get(
                    "profiling_prefix", None
                )
                if profile_prefix is None:
                    profile_prefix = join(
                        cfg.output_dir,
                        model_name,
                        f"{model_name}_{runtime_name}_profile",
                    )
                else:
                    profile_prefix = os.path.abspath(os.path.join(cfg.output_dir, profile_prefix))
                profile_prefix_dir = os.path.dirname(profile_prefix)
                assert os.path.isdir(profile_prefix_dir), f"Profiling directory {profile_prefix_dir} not found"
                profile_files = os.listdir(profile_prefix_dir)
                assert len(profile_files) > 0, f"No profiling files found in {profile_prefix}"
    results = pd.read_csv(join(cfg.output_dir, DATA_FILE))
    assert not results.empty, "Results CSV is empty"
    for model_name, model_cfg in cfg.nvbenjo.models.items():
        assert model_name in results.model.to_numpy(), "Model column not found in results"
        custom_metric_keys = model_cfg.get("custom_batchmetrics", {})
        for key in custom_metric_keys:
            assert key in results.columns, f"Custom batch metric {key} not found in results"


def test_default():
    with initialize(version_base=None, config_path="conf"):
        with tempfile.TemporaryDirectory() as tmpdir:
            cfg = compose(config_name="default", overrides=[f"output_dir={tmpdir}"])
            run_config(cfg)


def test_small_single():
    with initialize(version_base=None, config_path="conf"):
        with tempfile.TemporaryDirectory() as tmpdir:
            cfg = compose(config_name="small_single", overrides=[f"output_dir={tmpdir}"])
            run_config(cfg)
            _check_run_files(cfg)


def test_min_max_input_type():
    with initialize(version_base=None, config_path="conf"):
        with tempfile.TemporaryDirectory() as tmpdir:
            cfg = compose(config_name="input_min_max", overrides=[f"output_dir={tmpdir}"])
            with pytest.raises(ValueError):
                run_config(cfg)


class DummyModel(torch.nn.Module):
    def __init__(self):
        super(DummyModel, self).__init__()
        self.fc = torch.nn.Linear(10, 2)

    def forward(self, x):
        return self.fc(x)


def test_torch_load():
    model = DummyModel()

    with initialize(version_base=None, config_path="conf"):
        with tempfile.NamedTemporaryFile(delete=False, suffix=".pt") as tmpfile:
            torch.save(model, tmpfile)
            with tempfile.TemporaryDirectory() as tmpoutdir:
                cfg = compose(
                    config_name="torch_load",
                    overrides=[
                        f"output_dir={tmpoutdir}",
                        f'nvbenjo.models.dummytorchmodel.type_or_path="{tmpfile.name}"',
                    ],
                )
                run_config(cfg)
                _check_run_files(cfg)


class DummyModelMultiInput(torch.nn.Module):
    def __init__(self):
        super(DummyModelMultiInput, self).__init__()
        self.fc1 = torch.nn.Linear(10, 1)
        self.fc2 = torch.nn.Linear(20, 1)

    def forward(self, x, y):
        return self.fc1(x) * self.fc2(y)


def test_torch_load_multiinput():
    model = DummyModelMultiInput()

    with initialize(version_base=None, config_path="conf"):
        with tempfile.NamedTemporaryFile(delete=False, suffix=".pt") as tmpfile:
            torch.save(model, tmpfile)
            with tempfile.TemporaryDirectory() as tmpoutdir:
                cfg = compose(
                    config_name="torch_load_multiinput",
                    overrides=[
                        f"output_dir={tmpoutdir}",
                        f'nvbenjo.models.dummytorchmodel.type_or_path="{tmpfile.name}"',
                    ],
                )
                run_config(cfg)
                _check_run_files(cfg)


class ComplexDummyModelMultiInput(torch.nn.Module):
    def __init__(self, min, max):
        super(ComplexDummyModelMultiInput, self).__init__()
        self.fc1 = torch.nn.Linear(10, 1)
        self.fc2 = torch.nn.Linear(20, 1)
        self.min = min
        self.max = max

    def forward(self, x, y):
        if torch.any(x < self.min) or torch.any(x > self.max):
            raise ValueError(f"Input x contains values outside the range [{self.min}, {self.max}]")
        return self.fc1(x) * self.fc2(y)


@pytest.mark.parametrize("export_type", ["aot", "torchexport", "torchsave", "torchscript"])
@pytest.mark.parametrize("input_style", ["args", "kwargs"])
def test_torch_load_complex_multiinput_export_types(export_type, input_style):
    if export_type == "torchexport" and torch.__version__ < "2.1":
        pytest.skip("torch.export is only available in PyTorch 2.1 and later")
    if export_type == "aot" and torch.__version__ < "2.6":
        pytest.skip("aoti_compile_and_package is only available in PyTorch 2.6 and later")

    min = 12
    max = 34
    if input_style == "args":
        model = DummyModelMultiInput()
    else:
        model = ComplexDummyModelMultiInput(min=min, max=max)

    with initialize(version_base=None, config_path="conf"):
        suffix = ".pt2" if export_type in ["aot"] else ".pt"
        with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as tmpfile:
            if export_type == "torchsave":
                torch.save(model, tmpfile)
            elif export_type in ("torchexport", "aot"):
                # torchexport/aot need a model with simplified control flow
                model = DummyModelMultiInput()
                batch_size_dim = torch.export.Dim("B", min=1, max=1024)
                if input_style == "args":
                    program = torch.export.export(
                        model,
                        args=(torch.randn(2, 10), torch.randn(2, 20)),
                        dynamic_shapes={"x": {0: batch_size_dim}, "y": {0: batch_size_dim}},
                    )
                else:
                    program = torch.export.export(
                        model,
                        args=(),
                        kwargs={"x": torch.randn(2, 10), "y": torch.randn(2, 20)},
                        dynamic_shapes={"x": {0: batch_size_dim}, "y": {0: batch_size_dim}},
                    )
                if export_type == "torchexport":
                    torch.export.save(program, tmpfile.name)
                    torch.export.load(tmpfile.name)  # verify it can be loaded
                elif export_type == "aot":
                    torch._inductor.aoti_compile_and_package(
                        program,
                        package_path=tmpfile.name,
                    )
                else:
                    raise ValueError("Invalid export type")
            elif export_type == "torchscript":
                m = torch.jit.script(model)
                torch.jit.save(m, tmpfile)
            else:
                raise ValueError(f"Unknown export_type {export_type}")

            with tempfile.TemporaryDirectory() as tmpoutdir:
                config_override = {
                    "nvbenjo": {
                        "models": {
                            "dummytorchmodel": {
                                "type_or_path": tmpfile.name,
                                "num_batches": 2,
                                "batch_sizes": [1, 2],
                                "devices": ["cpu"],
                                "runtime_options": {
                                    "FP32": {"precision": "FP32", "compile": False},
                                    "FP16": {"precision": "FP16", "compile": False},
                                    "AMP_FP16": {"precision": "AMP_FP16", "compile": False},
                                },
                                "shape": [["B", 10], ["B", 20]]
                                if input_style == "args"
                                else [
                                    {"name": "x", "shape": ["B", 10], "min_max": [min, max]},
                                    {"name": "y", "shape": ["B", 20], "min_max": [min, max]},
                                ],
                            }
                        }
                    }
                }
                if torch.__version__ < "2.2":
                    config_override["nvbenjo"]["models"]["dummytorchmodel"]["runtime_options"].pop("AMP_FP16", None)
                    config_override["nvbenjo"]["models"]["dummytorchmodel"]["runtime_options"].pop("FP16", None)

                if export_type == "aot":
                    config_override["nvbenjo"]["models"]["dummytorchmodel"]["type_or_path"] = f"aot:{tmpfile.name}"
                    # AOT models are locked to their compiled precision
                    config_override["nvbenjo"]["models"]["dummytorchmodel"]["runtime_options"] = {
                        "FP32": {"precision": "FP32", "compile": False},
                    }
                cfg = compose(
                    config_name="default",
                    overrides=[
                        f"output_dir={tmpoutdir}",
                    ],
                )

                # temporary disable struct mode to allow merging additional model
                omegaconf.OmegaConf.set_struct(cfg, False)
                cfg = omegaconf.OmegaConf.merge(cfg, omegaconf.OmegaConf.create(config_override))
                omegaconf.OmegaConf.set_struct(cfg, True)
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    run_config(cfg)
                _check_run_files(cfg)


def test_torch_load_complex_invalid_multiinput():
    min = 0
    max = 12
    model = ComplexDummyModelMultiInput(min=min, max=max)

    with initialize(version_base=None, config_path="conf"):
        with tempfile.NamedTemporaryFile(delete=False, suffix=".pt") as tmpfile:
            torch.save(model, tmpfile)
            with tempfile.TemporaryDirectory() as tmpoutdir:
                config_override = {
                    "nvbenjo": {
                        "models": {
                            "dummytorchmodel": {
                                "type_or_path": tmpfile.name,
                                "num_batches": 2,
                                "batch_sizes": [1, 2],
                                "devices": ["cpu"],
                                "runtime_options": {
                                    "FP32": {
                                        "precision": "FP32",
                                        "compile": False,
                                        "enable_profiling": True,
                                        "profiler_kwargs": {"record_shapes": True, "with_stack": True},
                                    },
                                },
                                "shape": [
                                    {"name": "x", "shape": ["B", 10], "type": "float", "min_max": [max, max * 2]},
                                    {"name": "y", "shape": ["B", 20], "type": "float", "min_max": [max, max * 2]},
                                ],
                            }
                        }
                    }
                }
                cfg = compose(
                    config_name="default",
                    overrides=[
                        f"output_dir={tmpoutdir}",
                    ],
                )
                # temporary disable struct mode to allow merging additional model
                omegaconf.OmegaConf.set_struct(cfg, False)
                cfg = omegaconf.OmegaConf.merge(cfg, omegaconf.OmegaConf.create(config_override))
                omegaconf.OmegaConf.set_struct(cfg, True)
                if isinstance(cfg, omegaconf.DictConfig):
                    with pytest.raises(
                        ValueError, match=f"Input x contains values outside the range \\[{min}, {max}\\]"
                    ):
                        run_config(cfg)
                else:
                    raise ValueError("Config is not a DictConfig instance")


@pytest.mark.parametrize("compile_mode", ["aot_compile", "torch_compile"])
@pytest.mark.parametrize("precision", ["FP32", "FP16", "AMP_FP16"])
@pytest.mark.parametrize("extra_compile_kwargs", [False, True], ids=["default_kwargs", "extra_kwargs"])
@pytest.mark.parametrize(
    "model_cls,shape",
    [
        (DummyModel, ["B", 10]),
        (DummyModelMultiInput, [["B", 10], ["B", 20]]),
        (
            DummyModelMultiInput,
            [
                {"name": "x", "shape": ["B", 10]},
                {"name": "y", "shape": ["B", 20]},
            ],
        ),
    ],
    ids=["single", "multi_args", "multi_kwargs"],
)
def test_compile_modes(compile_mode, precision, extra_compile_kwargs, model_cls, shape):
    if compile_mode == "aot_compile" and torch.__version__ < "2.6":
        pytest.skip("aoti_compile_and_package is only available in PyTorch 2.6 and later")
    if precision in ("FP16", "AMP_FP16") and torch.__version__ < "2.2":
        pytest.skip("FP16/AMP_FP16 requires PyTorch 2.2 and later")
    if compile_mode == "aot_compile" and precision == "AMP_FP16":
        pytest.skip("AOT compiled models cannot use AMP precision")

    if extra_compile_kwargs:
        if compile_mode == "aot_compile":
            compile_kwargs = {"inductor_configs": {}}
        else:
            compile_kwargs = {"dynamic": True}
    else:
        compile_kwargs = {}

    model = model_cls()

    with initialize(version_base=None, config_path="conf"):
        with tempfile.NamedTemporaryFile(delete=False, suffix=".pt") as tmpfile:
            torch.save(model, tmpfile)
            with tempfile.TemporaryDirectory() as tmpoutdir:
                config_override = {
                    "nvbenjo": {
                        "models": {
                            "dummytorchmodel": {
                                "type_or_path": tmpfile.name,
                                "num_batches": 2,
                                "batch_sizes": [1],
                                "devices": ["cpu"],
                                "runtime_options": {
                                    precision: {
                                        "precision": precision,
                                        "compile": compile_mode,
                                        "compile_kwargs": compile_kwargs,
                                    },
                                },
                                "shape": shape,
                            }
                        }
                    }
                }
                cfg = compose(
                    config_name="default",
                    overrides=[f"output_dir={tmpoutdir}"],
                )
                omegaconf.OmegaConf.set_struct(cfg, False)
                cfg = omegaconf.OmegaConf.merge(cfg, omegaconf.OmegaConf.create(config_override))
                omegaconf.OmegaConf.set_struct(cfg, True)
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    run_config(cfg)
                _check_run_files(cfg)


@pytest.mark.skipif(torch.__version__ < "2.6", reason="aoti_compile_and_package requires PyTorch 2.6+")
def test_aot_prefix_loading():
    model = DummyModelMultiInput()
    batch_size_dim = torch.export.Dim("B", min=1, max=1024)
    program = torch.export.export(
        model,
        args=(torch.randn(2, 10), torch.randn(2, 20)),
        dynamic_shapes={"x": {0: batch_size_dim}, "y": {0: batch_size_dim}},
    )

    with initialize(version_base=None, config_path="conf"):
        with tempfile.NamedTemporaryFile(delete=False, suffix=".pt2") as tmpfile:
            torch._inductor.aoti_compile_and_package(
                program,
                package_path=tmpfile.name,
            )
            with tempfile.TemporaryDirectory() as tmpoutdir:
                config_override = {
                    "nvbenjo": {
                        "models": {
                            "aotmodel": {
                                "type_or_path": f"aot:{tmpfile.name}",
                                "num_batches": 2,
                                "batch_sizes": [1, 2],
                                "devices": ["cpu"],
                                "runtime_options": {
                                    "FP32": {"precision": "FP32", "compile": False},
                                },
                                "shape": [["B", 10], ["B", 20]],
                            }
                        }
                    }
                }
                cfg = compose(
                    config_name="default",
                    overrides=[f"output_dir={tmpoutdir}"],
                )
                omegaconf.OmegaConf.set_struct(cfg, False)
                cfg = omegaconf.OmegaConf.merge(cfg, omegaconf.OmegaConf.create(config_override))
                omegaconf.OmegaConf.set_struct(cfg, True)
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    run_config(cfg)
                _check_run_files(cfg)


def test_cli_cn_path_arg():
    with initialize(version_base=None, config_path="conf"):
        with tempfile.TemporaryDirectory() as cfg_tmpdir:
            cfg_file = os.path.join(cfg_tmpdir, "smallasdf.yaml")
            shutil.copy2(os.path.join("tests", "conf", "small_single.yaml"), cfg_file)
            with tempfile.TemporaryDirectory() as tmpdir:
                subprocess.run(
                    [
                        "python",
                        "-m",
                        "nvbenjo.cli",
                        "-cn",
                        cfg_file,
                        f"output_dir={tmpdir}",
                    ],
                    check=True,
                )

                cfg = compose(
                    config_name="small_single",
                    overrides=[
                        f"output_dir={tmpdir}",
                    ],
                )
                _check_run_files(cfg)



import os
import omegaconf
import pytest
import tempfile

import torch
from hydra import compose, initialize

from tests.test_cli import DummyModelMultiInput, _check_run_files, run_config

try:
    from nvbenjo import onnx_utils  # noqa: F401
except ImportError as e:
    if "onnxruntime" in str(e):
        pytest.skip("onnxruntime is not installed, skipping ONNX utils tests.", allow_module_level=True)
    else:
        raise


@pytest.mark.parametrize("do_profile", [True, False])
def test_onnx(do_profile: bool):
    model = DummyModelMultiInput()

    with initialize(version_base=None, config_path="conf"):
        with tempfile.NamedTemporaryFile(delete=False, suffix=".onnx") as tmpfile:
            torch.onnx.export(
                model,
                args=(torch.randn(2, 10), torch.randn(2, 20)),
                dynamic_axes={
                    "x": {0: "batch_size"},  # First dimension of input 'x' is dynamic
                    "y": {0: "batch_size"},  # First dimension of input 'y' is dynamic
                    "output": {0: "batch_size"},  # First dimension of output is dynamic
                },
                f=tmpfile.name,
                input_names=["x", "y"],
                output_names=["output"],
                opset_version=17,
            )
            min, max = 0, 5

            with tempfile.TemporaryDirectory() as tmpoutdir:
                profile_prefix_dir = "myprefix"
                profile_fileprefix = "testfile"
                profile_prefix = f"{profile_prefix_dir}/{profile_fileprefix}" if do_profile else None
                config_override = {
                    "nvbenjo": {
                        "models": {
                            "dummyonnxmodel": {
                                "type_or_path": tmpfile.name,
                                "num_batches": 2,
                                "batch_sizes": [1, 2],
                                "devices": ["cpu"],
                                "shape": [
                                    {"name": "x", "shape": ["B", 10], "min_max": [min, max]},
                                    {"name": "y", "shape": ["B", 20], "min_max": [min, max]},
                                ],
                                "runtime_options": {
                                    "default": {
                                        "graph_optimization_level": "ORT_ENABLE_ALL",
                                        "intra_op_num_threads": 1,
                                        "inter_op_num_threads": 2,
                                        "enable_profiling": do_profile,
                                        "profiling_prefix": profile_prefix,
                                    }
                                },
                            }
                        }
                    }
                }
                cfg = compose(
                    config_name="default",
                    overrides=[
                        f"output_dir={tmpoutdir}",
                    ],
                )
                # temporary disable struct mode to allow merging additional model
                omegaconf.OmegaConf.set_struct(cfg, False)
                cfg = omegaconf.OmegaConf.merge(cfg, omegaconf.OmegaConf.create(config_override))
                omegaconf.OmegaConf.set_struct(cfg, True)
                run_config(cfg)
                _check_run_files(cfg)
                assert os.path.isdir(os.path.join(tmpoutdir, profile_prefix_dir)) == do_profile
                if do_profile:
                    listprofile = os.listdir(os.path.join(tmpoutdir, profile_prefix_dir))
                    assert len(listprofile) > 0
                    for f in listprofile:
                        assert f.startswith(profile_fileprefix)



from unittest.mock import patch

import pandas as pd
import pytest

from nvbenjo.plot import print_system_info, visualize_results


@pytest.fixture
def sample_results():
    return pd.DataFrame(
        {
            "model": ["m1", "m1", "m2", "m2"],
            "device": ["cpu", "cpu", "cpu", "cpu"],
            "batch_size": [1, 1, 1, 1],
            "runtime_options": ["opt1", "opt1", "opt1", "opt1"],
            "time_inference": [0.1, 0.2, 0.3, 0.4],
        }
    )


@patch("nvbenjo.plot.plt")
@patch("nvbenjo.plot.sns")
def test_visualize_results_summary_plot(mock_sns, mock_plt, sample_results, tmp_path):
    """Multiple models, single device triggers the summary plot (lines 55-71)."""
    visualize_results(sample_results, str(tmp_path), keys=["time_inference"])
    assert mock_plt.savefig.call_count == 3  # 2 per-model + 1 summary


@patch("nvbenjo.plot.plt")
@patch("nvbenjo.plot.sns")
def test_visualize_results_empty_model_device(mock_sns, mock_plt, tmp_path):
    """Model+device combo with no rows triggers continue (line 36)."""
    results = pd.DataFrame(
        {
            "model": ["m1"],
            "device": ["cpu"],
            "batch_size": [1],
            "runtime_options": ["opt1"],
            "time_inference": [0.1],
        }
    )
    # Two devices in data but m1 only has cpu rows → cuda combo is empty
    results = pd.concat(
        [
            results,
            pd.DataFrame(
                {
                    "model": ["m2"],
                    "device": ["cuda"],
                    "batch_size": [1],
                    "runtime_options": ["opt1"],
                    "time_inference": [0.2],
                }
            ),
        ]
    )
    visualize_results(results, str(tmp_path), keys=["time_inference"])
    # m1+cuda and m2+cpu are empty → continue. Only m1+cpu and m2+cuda produce plots.
    assert mock_plt.savefig.call_count == 2


@patch("nvbenjo.plot.console")
def test_print_system_info_no_gpus(mock_console):
    """No GPUs triggers the 'None' branch (line 106)."""
    system_info = {
        "os": {"system": "Linux", "node": "test", "release": "6.0", "version": "#1"},
        "cpu": {"model": "TestCPU", "architecture": "x86_64", "cores": {"physical": 4, "total": 8}},
        "gpus": [],
    }
    print_system_info(system_info)
    mock_console.print.assert_called_once()



import os
from os.path import join
from typing import List

import matplotlib.pyplot as plt
import pandas as pd
from rich import box
import seaborn as sns
from rich.bar import Bar
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from . import console
from .utils import format_num, format_seconds


def visualize_results(
    results: pd.DataFrame,
    output_dir: str,
    keys: List[str] = [
        "time_cpu_to_device",
        "time_device_to_cpu",
        "time_inference",
        "time_total_batch_normalized",
        "memory_bytes",
    ],
    hue="runtime_options",
    col="batch_size",
    kind="bar",
):
    sns.set_style("whitegrid")
    for model in results.model.unique():
        mult_devices = len(results.device.unique()) > 1
        for device in results.device.unique():
            model_device_results = results[(results.model == model) & (results.device == device)]
            if len(model_device_results) == 0:
                continue
            for key in keys:
                if key in model_device_results.columns and not model_device_results[key].isnull().all():
                    sns.catplot(
                        data=model_device_results,
                        x="model",
                        y=key,
                        hue=hue,
                        col=col,
                        kind=kind,
                        palette="dark",
                        alpha=0.6,
                    )
                    device_stem = f"{device}_" if mult_devices else ""
                    os.makedirs(join(output_dir, model), exist_ok=True)
                    plt.savefig(join(output_dir, model, f"{device_stem}{key}.png"))
                    plt.close()

    if len(results.device.unique()) == 1 and len(results.model.unique()) > 1:
        for key in keys:
            if key in results.columns and not results[key].isnull().all():
                sns.catplot(
                    data=results.drop_duplicates().reset_index(drop=True),
                    y=key,
                    hue=hue,
                    col=col,
                    kind=kind,
                    row="model",
                    sharey=True,
                    palette="dark",
                    alpha=0.6,
                )
                device_stem = f"{device}_" if mult_devices else ""
                os.makedirs(join(output_dir, "summary"), exist_ok=True)
                plt.savefig(join(output_dir, "summary", f"summary_{key}.png"))
                plt.close()

    # TODO: maybe also check if only one model type then do same for device


def print_system_info(system_info: dict):
    text_color = "white"
    os_info = system_info["os"]
    os_string = os_info["system"].replace("Linux", "Linux 🐧")
    cpu_info = system_info["cpu"]
    gpu_infos = system_info["gpus"]
    driver_version = set(gpu_info["driver"] for gpu_info in gpu_infos)
    driver_version = driver_version.pop() if len(driver_version) == 1 else driver_version

    title = Text("System Information", style="bold cyan")

    content = Text()
    content.append("\n")
    content.append(f"💻️ {os_info['node']}\n", style="bold")
    content.append("OS:   ", style="bold yellow")
    content.append(f"{os_string} - {os_info['version']} ({os_info['release']})\n", style=text_color)
    content.append("CPU:  ", style="bold magenta")
    content.append(f"{cpu_info['model']} ({cpu_info['architecture']})\t", style=text_color)
    content.append("Cores: ", style=f"{text_color} bold")
    content.append(f"{cpu_info['cores']}\n", style=text_color)

    content.append("GPUs", style="bold green")
    if len(gpu_infos) > 0:
        content.append(f" (Driver {driver_version})\n", style="green")
        for gpu_info in gpu_infos:
            content.append("   ", style="bold blue")
            content.append(f"{gpu_info['name']} @ {gpu_info['clock_gpu']} ", style=text_color)
            content.append(f"({gpu_info['memory']} @ {gpu_info['clock_mem']})", style=text_color)
            content.append(f" - {gpu_info['architecture']}\n", style=text_color)
    else:
        content.append("  None\n", style=text_color)

    console.print(Panel(content, title=title, border_style="blue", padding=(0, 2)))


def _print_device_results(model_results: pd.Series | pd.DataFrame, model: str, device: str, custom_metric_keys: List):
    # Create a rich table for each model+device combination
    table = Table(
        title=f"Model: {model} on Device: {device}",
        show_header=True,
        header_style="bold",
        show_lines=True,
        title_style="bold",
    )

    # Get grouped results
    device_results = model_results[model_results.device == device]
    device_results = device_results.drop(columns=["device"])
    device_results = device_results.groupby(["model", "runtime_options", "batch_size"]).mean()
    device_results["device"] = device  # Add device column back for display
    print_result = device_results.reset_index()

    # Remove columns where all values are None
    print_result = print_result.dropna(axis="columns", how="all")

    # Format values for display
    for column in print_result.columns:
        if column == "time_total_batch_normalized":
            top3 = print_result.time_total_batch_normalized.nsmallest(3).index
            print_result[column] = print_result[column].apply(format_seconds)
            for i, emoji in enumerate(["🥇", "🥈", "🥉"][: len(top3)]):
                print_result.loc[top3[i], column] = f"{emoji} {print_result.loc[top3[i], column]}"
        elif column.startswith("time"):
            print_result[column] = print_result[column].apply(format_seconds)
        elif "bytes" in column:
            print_result[column] = print_result[column].apply(format_num, bytes=True)
        elif column == "device":
            print_result[column] = print_result[column].apply(lambda x: f"{x}")
        elif column in custom_metric_keys:
            print_result[column] = print_result[column].apply(lambda x: f"{x:.3f}")

    # Add columns to the table
    for col in print_result.columns:
        # Set column styles based on data type
        if col == "model":
            style = "bold green"
        elif col == "runtime_options":
            style = "bold blue"
        elif col == "batch_size":
            style = "bold yellow"
        elif col == "time_total_batch_normalized":
            style = "bold cyan"
        elif col.startswith("time"):
            style = None
        elif "memory" in col:
            style = "red"
        else:
            style = None

        # Format column names for better display
        display_name = col.replace("_", " ").title()
        table.add_column(display_name, style=style, justify="right")

    # Add rows to the table
    for _, row in print_result.iterrows():
        table.add_row(*[str(value) for value in row.values])

    # Display the table in a panel
    console.print(Panel(table, border_style="dim", padding=(0, 1)))


def _print_summary_plot(results: pd.Series | pd.DataFrame, custom_metric_keys: List):
    default_metric = "time_total_batch_normalized"
    default_metric_title = "Time Batch Normalized"
    if not custom_metric_keys:
        first_metric = default_metric
        metric_title = default_metric_title
    else:
        first_metric = custom_metric_keys[0]
        metric_title = first_metric

    table = Table(
        show_header=True,
        header_style="bold",
        box=box.SIMPLE,
        padding=(0, 1),
        title_justify="left",
        title_style="bold",
    )
    table.add_column("Model", style="bold")
    table.add_column("Runtime Options", style="bold")
    table.add_column("Batch Size", style="bold")
    table.add_column("Device", style="bold")
    table.add_column(metric_title, header_style="bold cyan")
    table.add_column("", justify="right")
    table.add_column("", justify="right")
    table.add_column("Memory", header_style="bold red")
    table.add_column("", justify="right")

    max_first_metric = results[first_metric].max().item()
    max_mem = results.memory_bytes.max().item()
    for model in results.model.unique():
        model_results = results[results.model == model]
        for device in model_results.device.unique():
            # Get grouped results and mean over runs
            device_results = model_results[model_results.device == device]
            device_results = device_results.drop(columns=["device"])
            device_results = device_results.groupby(["model", "runtime_options", "batch_size"]).mean()
            device_results["device"] = device  # Add device column back for display
            print_result = device_results.reset_index()

            # Remove columns where all values are None
            print_result = print_result.dropna(axis="columns", how="all")

            # Get grouped results
            device_results = model_results[model_results.device == device]

            print_result = print_result.sort_values(first_metric)
            for _, res in print_result.iterrows():
                first_val = res[first_metric]
                mem_val = res.memory_bytes
                table.add_row(
                    Text(model, style="green"),
                    Text(res.runtime_options, style="blue"),
                    Text(str(res.batch_size), style="yellow"),
                    res.device,
                    Bar(begin=0, size=max_first_metric, end=first_val, width=80, color="cyan"),
                    Text(
                        format_seconds(first_val) if first_metric == default_metric else str(format_num(first_val)),
                        style="cyan",
                    ),
                    "   ",
                    Bar(begin=0, size=max_mem, end=mem_val, width=80, color="red"),
                    Text(str(format_num(mem_val, bytes=True)), style="red"),
                )

    console.print(Panel(table, border_style="dim", padding=(0, 1)))


def print_results(results: pd.DataFrame, custom_metric_keys: List[str] = []):
    for model in results.model.unique():
        model_results = results[results.model == model]
        for device in model_results.device.unique():
            _print_device_results(model_results, model, device, custom_metric_keys)

    _print_summary_plot(results, custom_metric_keys)
